//
// PC-8001 USB CRT エミュレータ for Mac OS X
//			copyleft 2009 Koichi Nishida
//         on GNU General Public Licence

#include <Carbon/Carbon.h>
#include "WinTypes.h"
#include "ftd2xx.h"
#include "string.h"

// クリエーターコード
#define SIGNATURE 'P8US'

// 初期設定ファイル
#define PreferencesIdent CFSTR("P8USBCRT")

//　ダイアログの各コントロールのID
#define BUFSIZE_TEXT_ID 100
#define ROWBYTES_TEXT_ID 101
#define LINES_TEXT_ID 102
#define WIDE_CHECK_ID 103

// 画面の位置とサイズ
#define WIN_X 100
#define WIN_Y 100
#define WIN_WIDTH 430
#define WIN_HEIGHT 262

// タイトルの高さ
#define TITLE_HEIGHT 22

// USB のバッファサイズ
#define MAXBUFSIZE 256
int USBBufSize = 32;

// 一行のバイト数
int RowBytes = 416;

// ライン数
int Lines = 262;

// ウィンドウのハンドルとグラフィックポインタ
WindowRef window;
CGrafPtr gp;

// preferences ウインドウ
WindowRef prefWin;

// スクリーンの一行のバイト数とグラフィックメモリへのポインタ
unsigned int screenRowBytes;
unsigned long* XBuf;

// USB ハンドル
FT_HANDLE ftHandle;

//フレームバッファ
unsigned char *picbuf;

// wide モード & screen 4 カラーモード
int Wide = 1;

// プロトタイプ宣言
int InitUSB();
int GetUSB(unsigned char *dt);
void SetXBuf();

// 設定ファイルに int 値を設定
static void setPrefIntValue(CFStringRef key, int val)
{
	CFNumberRef aCFNumber = CFNumberCreate(kCFAllocatorDefault,
		kCFNumberSInt32Type, &val);
		
	CFPreferencesSetAppValue( key, aCFNumber, PreferencesIdent);
}

// 設定ファイルから int 値を取り出す
static Boolean getPrefIntValue(CFStringRef key, int* val)
{
	Boolean isValid = false;
	SInt32 intValue = CFPreferencesGetAppIntegerValue(
		key, PreferencesIdent, &isValid);
		
	if (isValid) *val = intValue;
	return isValid;
}

// タイマーのコールバック
static pascal void DisplayTimerFunc(EventLoopTimerRef timerref, void *userData)
{
	unsigned char dt = 0, prev, val, length, R = 0, G = 0, B = 0;
	int adr, ipic, i, x, y;
	static int bytes = 0, ucount = 0, tick = 0;
	unsigned long* p, *q, *r;
	unsigned int pval;

	// ヘッダ待ち
	do {
		prev = dt;
		if (GetUSB(&dt)) {
			// USB が抜けていたら終わる
			QuitApplicationEventLoop();
			return;
		}
		bytes++;
	} while (!((prev == 0 || prev == 1) && dt == 0x82));
	// ヘッダが	01 82	ならば I ピクチャ
	//			00 82	ならば P ピクチャ	
	ipic = prev;
RESET:
	dt = 0; val = 0; length = 0; adr = 0;
	p = XBuf;
	for (y = 0; y < Lines; y++) {
		q = p;
		r = q+screenRowBytes/4;
		// read a line
		for (x = 0; x < RowBytes; x++) {
			if (length == 0) {
				// 値と長さを得る
				if (GetUSB(&val)) {
					QuitApplicationEventLoop();
					return;
				}
				if (GetUSB(&length)) {
					QuitApplicationEventLoop();
					return;
				}
				bytes += 2;
				// if I picture header found, restart !
				if ((val == 0x01) && (length == 0x82)) { // ヘッダを検出したら。。。
					ipic = 1;
					goto RESET;
				}
			}
			// I ピクチャなら、値をそのまま使う
			// P ピクチャなら、前の値に足す		
			if (ipic) dt = val;
			else dt = picbuf[adr] + val;
			picbuf[adr++] = dt;
			length--;
			if (ipic || (val != 0)) {
				// 通常モード
				for (i=0; i<2; i++) {
					R = (dt&1)*255;
					G = ((dt>>1)&1)*255;
					B = ((dt>>2)&1)*255;
					pval = ((R << 16) | (G << 8) | B);
					if (i==0 || Wide) {
						*(p++) = pval;
					}
					if (Wide) {
						*(r++) = pval;
					}
					dt >>= 4;
				}
			} else {
				p += (Wide ? 2 : 1); r += (Wide ? 2 : 1);
			}
		}
		p = q + screenRowBytes/4*(Wide ? 2 : 1);
	}
	//　QuickDraw バッファをフラッシュ
	{
		Rect drt;
		drt.top = drt.left = 0;
		drt.bottom = Lines*(Wide+1);
		drt.right = RowBytes*(Wide+1); 
		RgnHandle rgn = NewRgn();
		RectRgn(rgn, &drt);    	
		QDFlushPortBuffer(gp, GetPortVisibleRegion(gp, rgn));
	}
	// タイトル表示
	if (++ucount == 5) {
		CFStringRef title;
		char cstr[64];
		int ticks = TickCount()-tick;
		tick = TickCount();
		sprintf(cstr, "P8USBCRT [%d frames/sec] [%d bytes/frame]", 300/ticks, bytes/5);
		title = CFStringCreateWithCString(NULL, cstr, kCFStringEncodingMacRoman);
		bytes = 0;
		ucount = 0;
		SetWindowTitleWithCFString(window, title);
		CFRelease(title);
	}

}

// テキストコントロールに数値を設定するローカル関数
static void SetTextIntItem(int ID, int data)
{
	ControlID       aCntrlID;
	ControlRef      aCntrlRef;
	CFStringRef string;
			
	aCntrlID.signature = SIGNATURE;
	aCntrlID.id = ID;
	GetControlByID(prefWin, &aCntrlID, &aCntrlRef);
	string = CFStringCreateWithFormat(NULL, NULL, CFSTR("%d"), data);
	SetControlData (aCntrlRef, kControlEntireControl,
		kControlEditTextCFStringTag, sizeof (CFStringRef), &string);
	DrawOneControl(aCntrlRef);
}

// テキストコントロールから数値を得るローカル関数
static int GetTextIntItem(int ID)
{
	ControlID       aCntrlID;
	ControlRef      aCntrlRef;
	CFStringRef string;
	int val;

	aCntrlID.signature = SIGNATURE;
	aCntrlID.id = ID;
	GetControlByID(prefWin, &aCntrlID, &aCntrlRef);
	GetControlData(aCntrlRef, 0, kControlEditTextCFStringTag, 
		sizeof(string), &string, NULL);
	val = CFStringGetIntValue(string);
	CFRelease(string);
	return val;
}

// メニューコマンドのコールバック
static pascal OSStatus DoAppCommandProcess(EventHandlerCallRef nextHandler, EventRef theEvent,
	void* userData)
{
	unsigned long ekind;
	long cls;
	HICommand aCommand;
	OSStatus result = eventNotHandledErr;

	cls = GetEventClass(theEvent);
	ekind = GetEventKind(theEvent);
	if (cls == kEventClassCommand) {
		GetEventParameter(theEvent, kEventParamDirectObject, typeHICommand,
			NULL, sizeof(HICommand), NULL, &aCommand);
		// menu
		switch (aCommand.commandID) {
		case kHICommandQuit:
			QuitApplicationEventLoop();
			result = noErr;
			break;
		case 'Wide': {
			Rect rect;
			Wide = !Wide;
			SetMenuCommandMark(NULL, 'Wide', Wide ? 0x0012 : 0);
			SetRect(&rect, WIN_X, WIN_Y, WIN_X+RowBytes*(Wide+1), WIN_Y+Lines*(Wide+1));
			SetWindowBounds(window, kWindowContentRgn, &rect);
			SetXBuf();
			result = noErr;
			break;
		}
		case 'pref':
		{   // 初期設定ダイアログのデータをセット
			static ControlID aCntrlID = {SIGNATURE, WIDE_CHECK_ID};
			ControlRef      aCntrlRef;
			ShowWindow(prefWin);
			GetControlByID(prefWin, &aCntrlID, &aCntrlRef);
			SetControl32BitValue(aCntrlRef, Wide);
			SetTextIntItem(BUFSIZE_TEXT_ID, USBBufSize);
			SetTextIntItem(ROWBYTES_TEXT_ID, RowBytes);
			SetTextIntItem(LINES_TEXT_ID, Lines);
				break;
		}
		}
	}
	return result;
}

// メインウインドウのサイズを変更
static void ResizeWin()
{
	Rect rect;
	
	SetRect(&rect, WIN_X, WIN_Y, WIN_X+RowBytes*(Wide+1), WIN_Y+Lines*(Wide+1));
	SetWindowBounds(window, kWindowContentRgn, &rect);
	SetXBuf();
}

// 初期設定ダイアログのコールバック
static pascal OSStatus DoPrefCommandProcess(EventHandlerCallRef nextHandler,
	EventRef theEvent, void* userData)
{
	unsigned long ekind;
	long cls;
	HICommand aCommand;
	OSStatus result = eventNotHandledErr;

	cls = GetEventClass(theEvent);
	ekind = GetEventKind(theEvent);
	if (cls == kEventClassCommand) {
		GetEventParameter(theEvent, kEventParamDirectObject, typeHICommand,
			NULL, sizeof(HICommand), NULL, &aCommand);
		// menu
		switch (aCommand.commandID) {
		case 'ok  ':
			// バッファサイズ変更
			USBBufSize = GetTextIntItem(BUFSIZE_TEXT_ID);
			FT_SetUSBParameters(ftHandle, USBBufSize, 0);
			// Row Bytes 変更
			RowBytes = GetTextIntItem(ROWBYTES_TEXT_ID);
			// Lines 変更
			Lines = GetTextIntItem(LINES_TEXT_ID);
			HideWindow(prefWin);
			{   // Wide を得る
				static ControlID aCntrlID = {SIGNATURE, WIDE_CHECK_ID};
				ControlRef      aCntrlRef;
				GetControlByID(prefWin, &aCntrlID, &aCntrlRef);
				Wide = GetControl32BitValue(aCntrlRef);
			}
			// Window サイズ変更
			ResizeWin();
			result = noErr;
			break;
		case 'not!':
			HideWindow(prefWin);
			result = noErr;
			break;
			}
	}
	return result;
}		

int main(int argc, char* argv[])
{
    IBNibRef nibRef;
    OSStatus err;
	EventTimerInterval time;
	EventTypeSpec AppEventTypes[] = {{kEventClassCommand, kEventCommandProcess}};
	EventTypeSpec PrefEventTypes[] = {{kEventClassCommand, kEventCommandProcess}};
	EventHandlerRef appHandlerRef, prefHandlerRef;
	EventLoopTimerRef DisplayTimerHandler;
	Rect rect;

	// 初期設定ファイル読み込み
	if (!getPrefIntValue(CFSTR("USBBufSize"), &USBBufSize))
		USBBufSize = 32;
	if (!getPrefIntValue(CFSTR("RowBytes"), &RowBytes))
		RowBytes = 416;
	if (!getPrefIntValue(CFSTR("Lines"), &Lines))
		Lines = 262;
	if (!getPrefIntValue(CFSTR("Wide"), &Wide))
		Wide = 1;

	// nib を得る
	err = CreateNibReference(CFSTR("main"), &nibRef);
    require_noerr( err, CantGetNibRef );
    
     //　nib からメニューを作成 
	err = SetMenuBarFromNib(nibRef, CFSTR("MenuBar"));
    require_noerr( err, CantSetMenuBar );

	SetMenuCommandMark(NULL, 'Wide', Wide ? 0x0012 : 0);

	// メニューイベントハンドラ設定
	err = InstallApplicationEventHandler(NewEventHandlerUPP(DoAppCommandProcess), 1,
		AppEventTypes, NULL, &appHandlerRef);
	require_noerr( err, CantCreateAppEvent );

	// タイマー設定
	time = 0.1 /60.0;
	err = InstallEventLoopTimer(GetCurrentEventLoop(), 0, time, NewEventLoopTimerUPP(DisplayTimerFunc),
		NULL, &DisplayTimerHandler);
	require_noerr( err, CantCreateTimer );

    // メインウィンドウ作成 
	SetRect(&rect, WIN_X, WIN_Y, WIN_X+RowBytes*(Wide+1), WIN_Y+Lines*(Wide+1));
	err = CreateNewWindow(kDocumentWindowClass,
		kWindowStandardHandlerAttribute | kWindowCollapseBoxAttribute, &rect, &window);
	require_noerr(err, CantCreateWindow);
	ShowWindow(window);

	// 初期設定ウインドウ
	err = CreateWindowFromNib(nibRef, CFSTR("Preferences"), &prefWin);
	require_noerr(err, CantCreatePrefWin);

	// 初期設定ウインドウのイベントハンドラインストール
	err = InstallWindowEventHandler(prefWin, NewEventHandlerUPP(DoPrefCommandProcess),
		1, PrefEventTypes, NULL, &prefHandlerRef);
	require_noerr( err, CantCreatePrefEvent );

	// フレームバッファ確保
	if (!(picbuf = (unsigned char*)malloc(0x20000))) goto CantMalloc;
	memset(picbuf, 0, 0x20000);
 
	// USB 初期化
	if (InitUSB()) goto InitUSBError;

	// グラフィックバッファ設定
	SetXBuf();

    // イベントループ
	RunApplicationEventLoop();

	// USB を閉じる
	FT_Close(ftHandle);

InitUSBError:
	free(picbuf);
CantMalloc:
	RemoveEventHandler(prefHandlerRef);
CantCreatePrefEvent:
	DisposeWindow(prefWin);
CantCreatePrefWin:
	DisposeWindow(window);
CantCreateWindow:
	RemoveEventLoopTimer(DisplayTimerHandler);
CantCreateTimer:
	RemoveEventHandler(appHandlerRef);
CantCreateAppEvent:
CantSetMenuBar:
    DisposeNibReference(nibRef);
CantGetNibRef:
	// 初期設定ファイル書き出し
	setPrefIntValue(CFSTR("USBBufSize"), USBBufSize);
	setPrefIntValue(CFSTR("RowBytes"), RowBytes);
	setPrefIntValue(CFSTR("Lines"), Lines);
	setPrefIntValue(CFSTR("Wide"), Wide);
	CFPreferencesAppSynchronize(PreferencesIdent);

	return err;
}

// USB 初期化
int InitUSB()
{
	FT_STATUS ftStatus;
	
	// USB をオープン
	ftStatus = FT_OpenEx("P8USBCRT", FT_OPEN_BY_DESCRIPTION, &ftHandle);
	if (ftStatus != FT_OK) return 1;

	// Output mode　設定
	ftStatus = FT_SetBitMode(ftHandle, 0x00, 0x00);
	if (ftStatus != FT_OK) return 1;

	// バッファサイズ設定
	ftStatus = FT_SetUSBParameters(ftHandle, USBBufSize, 0);
	if (ftStatus != FT_OK) return 1;
	
	//Time Out　設定
	ftStatus = FT_SetTimeouts(ftHandle, 3000, 3000);
	if (ftStatus != FT_OK) return 1;
	
	return 0;
}

// USB からデータを得る
int GetUSB(unsigned char *dt)
{
	static unsigned char buff[MAXBUFSIZE];
	static unsigned long bytesRead = 0;
	static unsigned int ptr = 0;
	FT_STATUS ftStatus = FT_OK;
		
	if (ptr == 0) {
		ftStatus = FT_Read(ftHandle, buff, USBBufSize, &bytesRead);
		if (bytesRead == 0 || ftStatus == FT_IO_ERROR) return 1;
	}
	*dt = buff[ptr];	
	if (++ptr == bytesRead) ptr = 0; 
//		printf("%x ", *dt); fflush(stdout);	
	return 0;
}

// グラフィックバッファ設定
void SetXBuf()
{
	PixMapHandle pm;

	gp = GetWindowPort(window);
	LockPortBits(gp);
	pm = GetGWorldPixMap(gp);
	screenRowBytes = GetPixRowBytes(pm);
	XBuf = (unsigned long*)GetPixBaseAddr(pm)+(screenRowBytes/4*TITLE_HEIGHT);
	UnlockPortBits(gp);
}



